from pathlib import Path

import pytest
from ape.api.networks import LOCAL_NETWORK_NAME

from ape_starknet import tokens as _tokens

all_tokens = pytest.mark.parametrize(
    "token",
    (
        "eth",
        "test_token",
        # "proxy_token",  # TODO: Fix local proxies
    ),
)


@pytest.fixture(scope="module")
def token_contract(config, account, token_initial_supply, project):
    project_path = Path(__file__).parent.parent / "projects" / "token"

    with config.using_project(project_path):
        yield project.get_contract("TestToken").deploy(
            123123, 321321, token_initial_supply, account.address
        )


@pytest.fixture(scope="module")
def proxy_token_contract(config, account, token_initial_supply, token_contract, project):
    project_path = Path(__file__).parent.parent / "projects" / "proxy"

    with config.using_project(project_path):
        contract = project.get_contract("Proxy").deploy(token_contract.address)
        _tokens.add_token("proxy_token", LOCAL_NETWORK_NAME, contract.address)
        return _tokens["proxy_token"]


@pytest.fixture(scope="module")
def tokens(token_contract, proxy_token_contract, provider, account):
    _tokens.add_token("test_token", LOCAL_NETWORK_NAME, token_contract.address)
    return _tokens


@all_tokens
def test_get_balance(tokens, account, token_initial_supply, token):
    if token == "eth":
        # Likely spent fees
        assert tokens.get_balance(account)
    else:
        assert tokens.get_balance(account, token=token) == token_initial_supply


@all_tokens
def test_transfer(tokens, account, second_account, token):
    initial_balance = tokens.get_balance(second_account.address, token=token)
    tokens.transfer(account.address, second_account.address, 10, token=token)
    assert tokens.get_balance(second_account.address, token=token) == initial_balance + 10


def test_large_transfer(tokens, account, second_account):
    initial_balance = tokens.get_balance(second_account.address, token="test_token")

    # Value large enough to properly test Uint256 logic
    balance_to_transfer = 2**128 + 1
    tokens.transfer(
        account.address, second_account.address, balance_to_transfer, token="test_token"
    )
    actual = tokens.get_balance(second_account.address, token="test_token")
    expected = initial_balance + balance_to_transfer
    assert actual == expected


def test_event_log_arguments(token_contract, account, second_account):
    amount0 = 10
    amount1 = 2**128 + 42
    receipt = token_contract.fire_events(second_account.address, amount0, amount1, sender=account)

    transfer_logs = list(receipt.decode_logs(token_contract.Transfer))
    assert len(transfer_logs) == 1
    log = transfer_logs[0]
    assert log.from_ == int(account.address, 16)
    assert log.to == int(second_account.address, 16)
    assert log.value == amount0

    mint_logs = list(receipt.decode_logs(token_contract.Mint))
    assert len(mint_logs) == 1
    log = mint_logs[0]
    assert log.sender == int(account.address, 16)
    assert log.amount0 == amount0
    assert log.amount1 == amount1
    assert log.to == int(second_account.address, 16)



import os
from typing import Any, Dict, Iterator, List, Optional, Union
from urllib.error import HTTPError
from urllib.parse import urlparse
from urllib.request import urlopen

import requests
from ape.api import BlockAPI, ProviderAPI, ReceiptAPI, SubprocessProvider, TransactionAPI
from ape.api.networks import LOCAL_NETWORK_NAME
from ape.contracts import ContractInstance
from ape.exceptions import ProviderNotConnectedError, TransactionError, VirtualMachineError
from ape.types import AddressType, BlockID, ContractLog
from ape.utils import DEFAULT_NUMBER_OF_TEST_ACCOUNTS, cached_property
from ethpm_types import ContractType
from ethpm_types.abi import EventABI
from starknet_py.net import Client as StarknetClient
from starknet_py.net.models import parse_address
from starkware.starknet.definitions.transaction_type import TransactionType
from starkware.starknet.services.api.feeder_gateway.response_objects import (
    DeclareSpecificInfo,
    DeploySpecificInfo,
    InvokeSpecificInfo,
    StarknetBlock,
)
from starkware.starkware_utils.error_handling import StarkErrorCode

from ape_starknet.config import DEFAULT_PORT, StarknetConfig
from ape_starknet.exceptions import StarknetEcosystemError, StarknetProviderError
from ape_starknet.tokens import TokenManager
from ape_starknet.transactions import (
    ContractDeclaration,
    InvokeFunctionTransaction,
    StarknetTransaction,
)
from ape_starknet.utils import (
    ALPHA_MAINNET_WL_DEPLOY_TOKEN_KEY,
    DEFAULT_ACCOUNT_SEED,
    PLUGIN_NAME,
    get_chain_id,
    get_dict_from_tx_info,
    get_virtual_machine_error,
    handle_client_errors,
)
from ape_starknet.utils.basemodel import StarknetBase


class StarknetProvider(SubprocessProvider, ProviderAPI, StarknetBase):
    """
    A Starknet provider.
    """

    # Gets set when 'connect()' is called.
    client: Optional[StarknetClient] = None
    token_manager: TokenManager = TokenManager()
    default_gas_cost: int = 0
    cached_code: Dict[int, Dict] = {}

    @property
    def process_name(self) -> str:
        return "starknet-devnet"

    @property
    def is_connected(self) -> bool:
        was_successful = False
        try:
            urlopen(self.uri)
            was_successful = True
        except HTTPError as err:
            was_successful = err.code == 404  # Task failed successfully
        except Exception:
            was_successful = False

        if was_successful and self.client is None:
            self.client = StarknetClient(self.uri, chain=self.chain_id)

        return was_successful

    @property
    def starknet_client(self) -> StarknetClient:
        if not self.is_connected:
            raise StarknetProviderError("Provider is not connected to Starknet.")

        return self.client

    def build_command(self) -> List[str]:
        parts = urlparse(self.uri)
        return [
            self.process_name,
            "--host",
            str(parts.hostname),
            "--port",
            str(parts.port),
            "--accounts",
            str(DEFAULT_NUMBER_OF_TEST_ACCOUNTS),
            "--seed",
            str(DEFAULT_ACCOUNT_SEED),
        ]

    @cached_property
    def plugin_config(self) -> StarknetConfig:
        return self.config_manager.get_config(PLUGIN_NAME) or StarknetConfig()  # type: ignore

    @cached_property
    def uri(self) -> str:
        network_config = self.plugin_config.provider.dict().get(self.network.name)
        if not network_config:
            raise StarknetProviderError(f"Unknown network '{self.network.name}'.")

        return network_config.get("uri") or f"http://127.0.0.1:{DEFAULT_PORT}"

    def connect(self):
        if self.network.name == LOCAL_NETWORK_NAME:
            # Behave like a 'SubprocessProvider'
            if not self.is_connected:
                super().connect()

            self.start()

        self.client = StarknetClient(self.uri, chain=self.chain_id)

    def disconnect(self):
        self.client = None
        super().disconnect()

    def update_settings(self, new_settings: dict):
        pass

    @property
    def chain_id(self) -> int:
        return get_chain_id(self.network.name).value

    @handle_client_errors
    def get_balance(self, address: AddressType) -> int:
        account = self.account_contracts[address]
        return self.token_manager.get_balance(account.address)

    @handle_client_errors
    def get_code(self, address: str) -> bytes:
        # NOTE: Always return truthy value for code so that ape core works properly
        return self.get_code_and_abi(address).get("bytecode", b"PROXY")

    @handle_client_errors
    def get_abi(self, address: str) -> List[Dict]:
        return self.get_code_and_abi(address)["abi"]

    @handle_client_errors
    def get_nonce(self, address: AddressType) -> int:
        # Check if passing a public-key address of a local account
        if address in self.account_contracts.public_key_addresses:
            contract_address = self.account_contracts.get_account(address).address
            if contract_address:
                address = contract_address

        checksum_address = self.starknet.decode_address(address)
        contract = self.chain_manager.contracts.instance_at(checksum_address)

        if not isinstance(contract, ContractInstance):
            raise StarknetProviderError(f"Account contract '{checksum_address}' not found.")

        return contract.get_nonce()

    @handle_client_errors
    def estimate_gas_cost(self, txn: TransactionAPI) -> int:
        if self.network.name == LOCAL_NETWORK_NAME:
            return self.default_gas_cost

        if not isinstance(txn, StarknetTransaction):
            raise StarknetEcosystemError(
                "Unable to estimate the gas cost for a non-Starknet transaction "
                "using Starknet provider."
            )

        starknet_object = txn.as_starknet_object()

        if not self.client:
            raise ProviderNotConnectedError()

        return self.client.estimate_fee_sync(starknet_object)

    @property
    def gas_price(self) -> int:
        """
        **NOTE**: Currently, the gas price is fixed to always be 100 gwei.
        """
        return self.conversion_manager.convert("100 gwei", int)

    @handle_client_errors
    def get_block(self, block_id: BlockID) -> BlockAPI:
        if isinstance(block_id, (int, str)) and len(str(block_id)) == 76:
            kwarg = "block_hash"
        elif block_id in ("pending", "latest"):
            kwarg = "block_number"
        elif isinstance(block_id, int):
            kwarg = "block_number"
            if block_id < 0:
                latest_block_number = self.get_block("latest").number
                block_id_int = latest_block_number + block_id + 1
                if block_id_int < 0:
                    raise ValueError(
                        f"Negative block number '{block_id_int}' results in non-existent block."
                    )

                block_id = block_id_int

        else:
            raise ValueError(f"Unsupported BlockID type '{type(block_id)}'.")

        block = self.starknet_client.get_block_sync(**{kwarg: block_id})
        return self.starknet.decode_block(block.dump())

    def _get_block(self, block_id: BlockID) -> StarknetBlock:
        kwarg = (
            "block_hash"
            if isinstance(block_id, (int, str)) and len(str(block_id)) == 76
            else "block_number"
        )
        return self.starknet_client.get_block_sync(**{kwarg: block_id})

    @handle_client_errors
    def send_call(self, txn: TransactionAPI) -> bytes:
        if not isinstance(txn, InvokeFunctionTransaction):
            type_str = f"{txn.type!r}" if isinstance(txn.type, bytes) else str(txn.type)
            raise StarknetProviderError(
                f"Transaction must be from an invocation. Received type {type_str}."
            )

        if not self.client:
            raise ProviderNotConnectedError()

        starknet_obj = txn.as_starknet_object()
        return self.client.call_contract_sync(starknet_obj)  # type: ignore

    @handle_client_errors
    def get_transaction(self, txn_hash: str) -> ReceiptAPI:
        self.starknet_client.wait_for_tx_sync(txn_hash)
        txn_info = self.starknet_client.get_transaction_sync(tx_hash=txn_hash).transaction
        receipt = self.starknet_client.get_transaction_receipt_sync(
            tx_hash=txn_info.transaction_hash
        )
        receipt_dict: Dict[str, Any] = {"provider": self, **vars(receipt)}
        receipt_dict = get_dict_from_tx_info(txn_info, **receipt_dict)
        return self.starknet.decode_receipt(receipt_dict)

    def get_transactions_by_block(self, block_id: BlockID) -> Iterator[TransactionAPI]:
        block = self._get_block(block_id)
        for txn_info in block.transactions:
            txn_dict = get_dict_from_tx_info(txn_info)
            yield self.starknet.create_transaction(**txn_dict)

    @handle_client_errors
    def send_transaction(self, txn: TransactionAPI, token: Optional[str] = None) -> ReceiptAPI:
        txn_info = self._send_transaction(txn, token=token)
        invoking = txn.type == TransactionType.INVOKE_FUNCTION

        if "code" in txn_info and txn_info["code"] != StarkErrorCode.TRANSACTION_RECEIVED.name:
            raise TransactionError(message="Transaction not received.")

        error = txn_info.get("error", {})
        if error:
            message = error.get("message", error)
            raise StarknetProviderError(message)

        txn_hash = txn_info["transaction_hash"]
        receipt = self.get_transaction(txn_hash)

        if invoking and isinstance(txn, InvokeFunctionTransaction):
            returndata = txn_info.get("result", [])
            receipt.returndata = returndata.copy()

            if txn.original_method_abi:
                # Use ABI before going through account contract
                abi = txn.original_method_abi
                return_data = returndata[1:]
            else:
                abi = txn.method_abi
                return_data = returndata

            return_value = self.starknet.decode_returndata(abi, return_data)
            receipt.return_value = return_value

        return receipt

    @handle_client_errors
    def _send_transaction(
        self, txn: TransactionAPI, token: Optional[str] = None
    ) -> Union[DeclareSpecificInfo, DeploySpecificInfo, InvokeSpecificInfo]:
        txn = self.prepare_transaction(txn)
        if not token and hasattr(txn, "token") and txn.token:  # type: ignore
            token = txn.token  # type: ignore
        else:
            token = os.environ.get(ALPHA_MAINNET_WL_DEPLOY_TOKEN_KEY)

        if not isinstance(txn, StarknetTransaction):
            raise StarknetEcosystemError(
                "Unable to send non-Starknet transaction using a Starknet provider."
            )

        starknet_txn = txn.as_starknet_object()
        return self.starknet_client.add_transaction_sync(starknet_txn, token=token)

    @handle_client_errors
    def get_contract_logs(
        self,
        address: Union[AddressType, List[AddressType]],
        abi: Union[EventABI, List[EventABI]],
        start_block: Optional[int] = None,
        stop_block: Optional[int] = None,
        block_page_size: Optional[int] = None,
        event_parameters: Optional[Dict] = None,
    ) -> Iterator[ContractLog]:
        raise NotImplementedError("TODO")

    @handle_client_errors
    def prepare_transaction(self, txn: TransactionAPI) -> TransactionAPI:
        if txn.type == TransactionType.INVOKE_FUNCTION and not txn.max_fee:
            txn.max_fee = self.estimate_gas_cost(txn)

        return txn

    def set_timestamp(self, new_timestamp: int):
        pending_timestamp = self.get_block("pending").timestamp
        seconds_to_increase = new_timestamp - pending_timestamp
        response = requests.post(
            url=f"{self.uri}/increase_time", json={"time": seconds_to_increase}
        )
        response.raise_for_status()
        response_data = response.json()
        if "timestamp_increased_by" not in response_data:
            raise StarknetProviderError(response_data)

    def get_virtual_machine_error(self, exception: Exception) -> VirtualMachineError:
        return get_virtual_machine_error(exception) or VirtualMachineError(base_err=exception)

    def get_code_and_abi(self, address: Union[str, AddressType, int]):
        address_int = parse_address(address)

        # Cache code for faster look-up
        if address_int not in self.cached_code:
            self.cached_code[address_int] = self.starknet_client.get_code_sync(address_int)

        return self.cached_code[address_int]

    @handle_client_errors
    def declare(self, contract_type: ContractType) -> ContractDeclaration:
        transaction = self.starknet.encode_contract_declaration(contract_type)
        return self.provider.send_transaction(transaction)


__all__ = ["StarknetProvider"]



import re
from typing import Any, Dict, Optional, Union

from ape.api.networks import LOCAL_NETWORK_NAME
from ape.exceptions import ApeException, ContractLogicError, OutOfGasError, VirtualMachineError
from ape.types import AddressType, RawAddress
from eth_typing import HexAddress, HexStr
from eth_utils import add_0x_prefix, is_text, remove_0x_prefix
from ethpm_types import ContractType
from hexbytes import HexBytes
from starknet_py.net.client import BadRequest
from starknet_py.net.models import TransactionType
from starknet_py.net.models.address import parse_address
from starknet_py.transaction_exceptions import TransactionRejectedError
from starkware.crypto.signature.fast_pedersen_hash import pedersen_hash
from starkware.crypto.signature.signature import get_random_private_key as get_random_pkey
from starkware.starknet.definitions.general_config import StarknetChainId
from starkware.starknet.services.api.contract_class import ContractClass
from starkware.starknet.services.api.feeder_gateway.response_objects import (
    DeclareSpecificInfo,
    DeploySpecificInfo,
    InvokeSpecificInfo,
)

from ape_starknet.exceptions import StarknetProviderError

PLUGIN_NAME = "starknet"
NETWORKS = {
    # chain_id, network_id
    "mainnet": (StarknetChainId.MAINNET.value, StarknetChainId.MAINNET.value),
    "testnet": (StarknetChainId.TESTNET.value, StarknetChainId.TESTNET.value),
}
_HEX_ADDRESS_REG_EXP = re.compile("(0x)?[0-9a-f]*", re.IGNORECASE | re.ASCII)
"""Same as from eth-utils except not limited length."""
ALPHA_MAINNET_WL_DEPLOY_TOKEN_KEY = "ALPHA_MAINNET_WL_DEPLOY_TOKEN"
DEFAULT_ACCOUNT_SEED = 13333337


def get_chain_id(network_id: Union[str, int]) -> StarknetChainId:
    if isinstance(network_id, int):
        return StarknetChainId(network_id)

    elif network_id == LOCAL_NETWORK_NAME:
        return StarknetChainId.TESTNET  # Use TESTNET chain ID for local network

    elif network_id not in NETWORKS:
        raise ValueError(f"Unknown network '{network_id}'.")

    return StarknetChainId(NETWORKS[network_id][0])


def to_checksum_address(address: RawAddress) -> AddressType:
    if isinstance(address, bytes):
        address = HexBytes(address).hex()

    address_int = parse_address(address)
    address_str = pad_hex_str(HexBytes(address_int).hex().lower())
    chars = [c for c in remove_0x_prefix(HexStr(address_str))]
    hashed = [b for b in HexBytes(pedersen_hash(0, address_int))]

    for i in range(0, len(chars), 2):
        try:
            if hashed[i >> 1] >> 4 >= 8:
                chars[i] = chars[i].upper()
            if (hashed[i >> 1] & 0x0F) >= 8:
                chars[i + 1] = chars[i + 1].upper()
        except IndexError:
            continue

    return AddressType(HexAddress(add_0x_prefix(HexStr("".join(chars)))))


def is_hex_address(value: Any) -> bool:
    return _HEX_ADDRESS_REG_EXP.fullmatch(value) is not None if is_text(value) else False


def is_checksum_address(value: Any) -> bool:
    if not is_text(value):
        return False

    if not is_hex_address(value):
        return False

    return value == to_checksum_address(value)


def handle_client_errors(f):
    def func(*args, **kwargs):
        try:
            result = f(*args, **kwargs)
            if isinstance(result, dict) and result.get("error"):
                message = result["error"].get("message") or "Transaction failed"
                raise StarknetProviderError(message)

            return result

        except BadRequest as err:
            msg = err.text if hasattr(err, "text") else str(err)
            raise StarknetProviderError(msg) from err
        except ApeException:
            # Don't catch ApeExceptions, let them raise as they would.
            raise

        except TransactionRejectedError as err:
            vm_error = get_virtual_machine_error(err)
            if vm_error:
                raise vm_error from err

            raise  # Original exception

    return func


def get_virtual_machine_error(err: Exception) -> Optional[VirtualMachineError]:
    err_msg = str(err)

    if "rejected" not in err_msg:
        return None

    elif "actual fee exceeded max fee" in err_msg.lower():
        return OutOfGasError()  # type: ignore

    if "Error message: " in err_msg:
        err_msg = err_msg.split("Error message: ")[-1]
        if "Error at pc=" in err_msg:
            err_msg = err_msg.split("Error at pc=")[0]
    elif "error_message=" in err_msg:
        err_msg = err_msg.split("error_message=")[-1].strip("'")

    # Fix escaping newline issue with error message.
    err_msg = err_msg.replace("\\n", "").strip()
    err_msg = err_msg.replace(
        "Transaction was rejected with following starknet error: ", ""
    ).strip()
    return ContractLogicError(revert_message=err_msg)


def get_dict_from_tx_info(
    txn_info: Union[DeploySpecificInfo, InvokeSpecificInfo], **extra_kwargs
) -> Dict:
    txn_dict = {**txn_info.dump(), **extra_kwargs}
    if isinstance(txn_info, DeploySpecificInfo):
        txn_dict["contract_address"] = to_checksum_address(txn_info.contract_address)
        txn_dict["max_fee"] = 0
        txn_dict["type"] = TransactionType.DEPLOY
    elif isinstance(txn_info, InvokeSpecificInfo):
        txn_dict["contract_address"] = to_checksum_address(txn_info.contract_address)

        if "events" in txn_dict:
            txn_dict["events"] = [vars(e) for e in txn_dict["events"]]

        txn_dict["max_fee"] = txn_dict["max_fee"]

        if "method_abi" in txn_dict:
            txn_dict["method_abi"] = txn_dict.get("method_abi")

        if "entry_point_selector" in txn_dict:
            txn_dict["entry_point_selector"] = txn_dict["entry_point_selector"]

        txn_dict["type"] = TransactionType.INVOKE_FUNCTION
    elif isinstance(txn_info, DeclareSpecificInfo):
        txn_dict["sender"] = to_checksum_address(txn_info.sender_address)
        txn_dict["type"] = TransactionType.DECLARE

    return txn_dict


def convert_contract_class_to_contract_type(contract_class: ContractClass):
    return ContractType.parse_obj(
        {
            "contractName": "Account",
            "sourceId": "openzeppelin.account.Account.cairo",
            "deploymentBytecode": {"bytecode": contract_class.serialize().hex()},
            "runtimeBytecode": {},
            "abi": contract_class.abi,
        }
    )


def get_random_private_key() -> str:
    private_key = HexBytes(get_random_pkey()).hex()
    return pad_hex_str(private_key)


def pad_hex_str(value: str, to_length: int = 66) -> str:
    val = value.replace("0x", "")
    actual_len = len(val)
    padding = "0" * (to_length - 2 - actual_len)
    return f"0x{padding}{val}"



import pytest
from hexbytes import HexBytes

from ape_starknet.utils import get_random_private_key


@pytest.mark.parametrize("iteration", range(10))
def test_get_random_private_key(iteration):
    pkey = get_random_private_key()
    assert len(pkey) == 66
    pkey_int = int(pkey, 16)
    pkey_back_to_str = HexBytes(pkey_int).hex()
    assert pkey_back_to_str.replace("0x", "") in pkey



def test_can_connect_to_testnet(networks):
    with networks.parse_network_choice("starknet:testnet") as provider:
        eth_contract = provider.tokens["eth"]

        # Make call
        eth_contract.balanceOf("0x348ef2b95e31269b4a1c019428723e3a33cd964f92ad866741f189b88be3bc0")



import time

import pytest
from starkware.starknet.public.abi import get_selector_from_name

from ape_starknet.utils import is_checksum_address


def test_get_nonce(provider, account, contract):
    initial_nonce = provider.get_nonce(account.address)

    # Transact to increase nonce
    contract.increase_balance(account.address, 123, sender=account)

    actual = provider.get_nonce(account.address)
    assert actual == initial_nonce + 1


def test_get_block(provider, contract):
    _ = contract  # Contract fixture used to increase blocks (since deploys happen)
    latest_block_0 = provider.get_block("latest")
    latest_block_1 = provider.get_block(-1)
    latest_block_2 = provider.get_block(latest_block_0.number)
    assert latest_block_0.hash == latest_block_1.hash == latest_block_2.hash


def test_get_negative_block_number(account, provider, contract):
    start_block = provider.get_block("latest").number
    for _ in range(3):
        # Mine 3 blocks
        contract.increase_balance(account.address, 12, sender=account)

    block = provider.get_block(-2)
    assert block.number == start_block + 2


def test_get_block_negative_number_resulting_less_than_zero(provider, contract):
    _ = contract  # Contract fixture used to increase blocks (since deploys happen)
    latest_block_number = provider.get_block("latest").number
    value = -1000000000000000
    with pytest.raises(ValueError) as err:
        provider.get_block(value)

    expected_block_number = latest_block_number + value + 1
    assert (
        str(err.value)
        == f"Negative block number '{expected_block_number}' results in non-existent block."
    )


def test_get_transactions_by_block(provider, account, contract):
    # Transact to create data.
    expected_value = 123
    contract.increase_balance(account.address, expected_value, sender=account)

    transactions = [t for t in provider.get_transactions_by_block("latest")]

    expected_chain_id = provider.chain_id
    expected_abi = [
        a for a in account.get_contract_type().mutable_methods if a.name == "__execute__"
    ][0]
    expected_nonce = account.nonce - 1
    assert len(transactions) == 1
    assert transactions[0].chain_id == expected_chain_id
    assert transactions[0].method_abi == expected_abi
    assert transactions[0].receiver == account.address
    assert transactions[0].value == 0
    assert is_checksum_address(transactions[0].receiver)
    expected_data = [
        1,
        provider.starknet.encode_address(contract.address),
        get_selector_from_name("increase_balance"),
        0,
        2,
        2,
        provider.starknet.encode_address(account.address),
        expected_value,
        expected_nonce,
    ]
    assert transactions[0].data == expected_data


def test_set_timestamp(provider, account, contract):
    start_time = provider.get_block("pending").timestamp
    provider.set_timestamp(start_time + 8600)
    contract.increase_balance(account.address, 123, sender=account)
    curr_time = time.time()
    assert pytest.approx(curr_time + 8600) == provider.get_block("latest").timestamp
